"""Utils."""

__all__ = []


def ROL32(X: int, count: int):
    """Rotate left for 32 bit number."""

    count &= 0x1f
    return ((X << count) | (X >> (32 - count))) & 0xffffffff


def bytes_to_int(b: bytes) -> int:
    return int.from_bytes(b, "big")


def int_to_bytes(i: int) -> bytes:
    """Convert integer to minimum number of bytes required to store its value."""

    return i.to_bytes((i.bit_length() + 7) >> 3, "big")



import math
from typing import Callable, Tuple, Type

from . import ellipticcurve as Ec
from . import errors
from .base import KEYXCHG_MODE, PC_MODE, Hash, SMCoreBase
from .sm3 import SM3
from .utils import bytes_to_int, int_to_bytes

__all__ = [
    "SM2",
    "PC_MODE",
    "KEYXCHG_MODE",
]

_ecdlp = Ec.ECDLP(
    0xFFFFFFFE_FFFFFFFF_FFFFFFFF_FFFFFFFF_FFFFFFFF_00000000_FFFFFFFF_FFFFFFFF,
    0xFFFFFFFE_FFFFFFFF_FFFFFFFF_FFFFFFFF_FFFFFFFF_00000000_FFFFFFFF_FFFFFFFC,
    0x28E9FA9E_9D9F5E34_4D5A9E4B_CF6509A7_F39789F5_15AB8F92_DDBCBD41_4D940E93,
    (0x32C4AE2C_1F198119_5F990446_6A39C994_8FE30BBF_F2660BE1_715A4589_334C74C7,
     0xBC3736A2_F4F6779C_59BDCEE3_6B692153_D0A9877C_C62A4740_02DF32E5_2139F0A0),
    0xFFFFFFFE_FFFFFFFF_FFFFFFFF_FFFFFFFF_7203DF6B_21C6052B_53BBF409_39D54123,
)


def point_to_bytes(P: Ec.EcPoint, mode: PC_MODE) -> bytes:
    """Convert point to bytes."""

    if P == _ecdlp.ec.INF:
        return b"\x00"

    etob = _ecdlp.fp.etob
    x, y = P

    if mode is PC_MODE.RAW:
        return b"\x04" + etob(x) + etob(y)
    elif mode is PC_MODE.COMPRESS:
        if y & 0x1:
            return b"\x03" + etob(x)
        else:
            return b"\x02" + etob(x)
    elif mode is PC_MODE.MIXED:
        if y & 0x1:
            return b"\x07" + etob(x) + etob(y)
        else:
            return b"\x06" + etob(x) + etob(y)
    else:
        raise TypeError(f"Invalid mode {mode}")


def bytes_to_point(b: bytes) -> Ec.EcPoint:
    """Convert bytes to point."""

    fp = _ecdlp.fp
    ec = _ecdlp.ec

    mode = b[0]
    if mode == 0x00:
        return ec.INF

    point = b[1:]
    x = fp.btoe(point[:fp.e_length])
    if mode == 0x04 or mode == 0x06 or mode == 0x07:
        return x, fp.btoe(point[fp.e_length:])
    elif mode == 0x02 or mode == 0x03:
        y = ec.get_y(x)
        if y is None:
            raise errors.PointNotOnCurveError((x, y))
        ylsb = y & 0x1
        if mode == 0x02 and ylsb or mode == 0x03 and not ylsb:
            return x, fp.neg(y)
        return x, y
    else:
        raise errors.InvalidPCError(mode)


class SM2Core(SMCoreBase):
    """SM2 Core Algorithms."""

    def __init__(self, ecdlp: Ec.ECDLP, hash_cls: Type[Hash], rnd_fn: Callable[[int], int] = None) -> None:
        """SM2 Core Algorithms.

        Args:
            ecdlp (ECDLP): ECDLP used in SM2.
            hash_cls (Type[Hash]): Hash class used in SM2.
            rnd_fn ((int) -> int): Random function used to generate k-bit random number, default to `secrets.randbits`.
        """

        super().__init__(hash_cls, rnd_fn)

        self.ecdlp = ecdlp

        # used in key exchange
        w = math.ceil(math.ceil(math.log2(self.ecdlp.fpn.p)) / 2) - 1
        self._2w = 1 << w
        self._2w_1 = self._2w - 1

    def generate_pk(self, sk: int) -> Ec.EcPoint:
        """Generate public key by secret key."""

        return self.ecdlp.kG(sk)

    def generate_keypair(self) -> Tuple[int, Ec.EcPoint]:
        """Generate key pair."""

        sk = self._randint(1, self.ecdlp.fpn.p - 2)
        return sk, self.generate_pk(sk)

    def verify_pk(self, pk: Ec.EcPoint) -> bool:
        """Verify if a public key is valid."""

        ec = self.ecdlp.ec

        if pk == ec.INF:
            return False

        if not ec.isvalid(pk):
            return False

        if ec.mul(self.ecdlp.fpn.p, pk) != ec.INF:
            return False

        return True

    def entity_info(self, uid: bytes, pk: Ec.EcPoint) -> bytes:
        """Generate other entity information bytes.

        Raises:
            DataOverflowError: ID length more than 8192 bytes.
        """

        ENTL = len(uid) << 3
        if ENTL.bit_length() >= 16:
            raise errors.DataOverflowError("ID", "8192 bytes")

        etob = self.ecdlp.fp.etob
        xP, yP = pk
        xG, yG = self.ecdlp.G

        Z = bytearray()
        Z.extend(ENTL.to_bytes(2, "big"))
        Z.extend(uid)
        Z.extend(etob(self.ecdlp.ec.a))
        Z.extend(etob(self.ecdlp.ec.b))
        Z.extend(etob(xG))
        Z.extend(etob(yG))
        Z.extend(etob(xP))
        Z.extend(etob(yP))

        return self._hash_fn(Z)

    def sign(self, message: bytes, sk: int, uid: bytes, pk: Ec.EcPoint = None) -> Tuple[int, int]:
        """Generate signature on the message.

        Args:
            message (bytes): Message to be signed.
            sk (int): Secret key.
            uid (bytes): User ID.
            pk (EcPoint): Public key

        Returns:
            int: r.
            int: s.
        """

        if pk is None:
            pk = self.generate_pk(sk)

        e = bytes_to_int(self._hash_fn(self.entity_info(uid, pk) + message))

        ecdlp = self.ecdlp
        fpn = self.ecdlp.fpn
        while True:
            k = self._randint(1, fpn.p - 1)
            x, _ = ecdlp.kG(k)

            r = fpn.add(e, x)
            if fpn.iszero(r) or fpn.iszero(fpn.add(r, k)):
                continue

            s = fpn.mul(fpn.sub(k, fpn.mul(r, sk)), fpn.inv(1 + sk))
            if fpn.iszero(s):
                continue

            return r, s

    def verify(self, message: bytes, r: int, s: int, uid: bytes, pk: Ec.EcPoint) -> bool:
        """Verify the signature on the message.

        Args:
            message (bytes): Message to be verified.
            r (int): r
            s (int): s
            uid (bytes): User ID.
            pk (EcPoint): Public key.

        Returns:
            bool: Whether OK.
        """

        ec = self.ecdlp.ec
        fpn = self.ecdlp.fpn

        if r < 1 or r > fpn.p - 1:
            return False

        if s < 1 or s > fpn.p - 1:
            return False

        t = fpn.add(r, s)
        if fpn.iszero(t):
            return False

        e = int.from_bytes(self._hash_fn(self.entity_info(uid, pk) + message), "big")

        x, _ = ec.add(self.ecdlp.kG(s), ec.mul(t, pk))
        if fpn.add(e, x) != r:
            return False

        return True

    def encrypt(self, plain: bytes, pk: Ec.EcPoint) -> Tuple[Ec.EcPoint, bytes, bytes]:
        """Encrypt.

        Args:
            plain (bytes): Plain text to be encrypted.
            pk (EcPoint): Public key.

        Returns:
            EcPoint: C1, kG point
            bytes: C2, cipher
            bytes: C3, hash value

        Raises:
            InfinitePointError: Infinite point encountered.

        The return order is `C1, C2, C3`, **NOT** `C1, C3, C2`.
        """

        ec = self.ecdlp.ec

        while True:
            k = self._randint(1, self.ecdlp.fpn.p - 1)
            x1, y1 = self.ecdlp.kG(k)  # C1

            if ec.mul(self.ecdlp.h, pk) == ec.INF:
                raise errors.InfinitePointError(f"Infinite point encountered, [0x{self.ecdlp.h:x}](0x{pk[0]:x}, 0x{pk[1]:x})")

            x2, y2 = ec.mul(k, pk)
            x2 = self.ecdlp.fp.etob(x2)
            y2 = self.ecdlp.fp.etob(y2)

            t = self._key_derivation_fn(x2 + y2, len(plain))
            if not any(t):
                continue

            C2 = bytes(map(lambda b1, b2: b1 ^ b2, plain, t))
            C3 = self._hash_fn(x2 + plain + y2)

            return (x1, y1), C2, C3

    def decrypt(self, C1: Ec.EcPoint, C2: bytes, C3: bytes, sk: int) -> bytes:
        """Decrypt.

        Args:
            C1 (EcPoint): kG point.
            C2 (bytes): Cipher.
            C3 (bytes): Hash value.
            sk (int): Secret key.

        Returns:
            bytes: Plain text.

        Raises:
            PointNotOnCurveError: Invalid C1 point, not on curve.
            InfinitePointError: Infinite point encountered.
            UnknownError: Zero bytes key stream.
            CheckFailedError: Incorrect hash value.
        """

        ec = self.ecdlp.ec

        if not ec.isvalid(C1):
            raise errors.PointNotOnCurveError(C1)

        if ec.mul(self.ecdlp.h, C1) == ec.INF:
            raise errors.InfinitePointError(f"Infinite point encountered, [0x{self.ecdlp.h:x}](0x{C1[0]:x}, 0x{C1[1]:x})")

        x2, y2 = ec.mul(sk, C1)
        x2 = self.ecdlp.fp.etob(x2)
        y2 = self.ecdlp.fp.etob(y2)

        t = self._key_derivation_fn(x2 + y2, len(C2))
        if not any(t):
            raise errors.UnknownError("Zero bytes key stream.")

        M = bytes(map(lambda b1, b2: b1 ^ b2, C2, t))

        if self._hash_fn(x2 + M + y2) != C3:
            raise errors.CheckFailedError("Incorrect hash value.")

        return M

    def _x_bar(self, x: int):
        """Used in key exchange."""

        return self._2w + (x & self._2w_1)

    def begin_key_exchange(self, sk: int) -> Tuple[Ec.EcPoint, int]:
        """Generate data to begin key exchange.

        Returns:
            EcPoint: Random point, [r]G, r in [1, n - 1]
            int: t
        """

        ecdlp = self.ecdlp
        fpn = ecdlp.fpn

        r = self._randint(1, fpn.p - 1)
        R = ecdlp.kG(r)
        t = fpn.add(sk, fpn.mul(self._x_bar(R[0]), r))

        return R, t

    def get_secret_point(self, t: int, R: Ec.EcPoint, pk: Ec.EcPoint) -> Ec.EcPoint:
        """Generate same secret point as another user.

        Args:
            t (int): Generated from `begin_key_exchange`
            R (EcPoint): Random point from another user.
            pk (EcPoint): Public key of another user.

        Returns:
            EcPoint: The same secret point as another user.

        Raises:
            PointNotOnCurveError: `R` not on curve..
            InfinitePointError: Secret point is infinite point.
        """

        ec = self.ecdlp.ec

        if not ec.isvalid(R):
            raise errors.PointNotOnCurveError(R)

        S = ec.mul(self.ecdlp.h * t, ec.add(pk, ec.mul(self._x_bar(R[0]), R)))

        if S == ec.INF:
            raise errors.InfinitePointError("Infinite point encountered.")

        return S

    def generate_skey(self, klen: int, S: Ec.EcPoint,
                      uid_init: bytes, pk_init: Ec.EcPoint,
                      uid_resp: bytes, pk_resp: Ec.EcPoint) -> bytes:
        """Generate secret key of klen bytes as same as another user.

        Args:
            klen (int): key length in bytes to generate.
            x (int): x of secret point.
            y (int): y of secret point.

            uid_init (bytes): User ID bytes of initiator.
            pk_init (EcPoint): public key of initiator.

            uid_resp (bytes): User ID bytes of responder.
            pk_resp (EcPoint): public key of responder.

        Returns:
            bytes: secret key of klen bytes.
        """

        x, y = S

        Z = bytearray()

        Z.extend(self.ecdlp.fp.etob(x))
        Z.extend(self.ecdlp.fp.etob(y))
        Z.extend(self.entity_info(uid_init, pk_init))
        Z.extend(self.entity_info(uid_resp, pk_resp))

        return self._key_derivation_fn(Z, klen)


class SM2:
    """SM2."""

    def __init__(self, sk: bytes = None, uid: bytes = None, pk: bytes = None, *,
                 rnd_fn: Callable[[int], int] = None, pc_mode: PC_MODE = PC_MODE.RAW) -> None:
        """SM2.

        Args:
            sk (bytes): Secret key.
            pk (bytes): Public key.
            uid (bytes): User ID.

            rnd_fn ((int) -> int): Random function used to generate k-bit random number, default to `secrets.randbits`.
            pc_mode (PC_MODE): Point compress mode used for generated data, no effects on the data to be parsed.
        """

        self._core = SM2Core(_ecdlp, SM3, rnd_fn)
        self._sk = bytes_to_int(sk) if sk else None
        self._pk = self._get_pk(pk)

        self._uid = uid
        self._pc_mode = pc_mode

    def _get_pk(self, pk: bytes) -> Ec.EcPoint:
        if pk:
            return bytes_to_point(pk)
        else:
            if self._sk:
                return self._core.generate_pk(self._sk)  # try generate public key
            else:
                return None

    @property
    def can_sign(self) -> bool:
        return bool(self._sk and self._uid)

    @property
    def can_verify(self) -> bool:
        return bool(self._pk and self._uid)

    @property
    def can_encrypt(self) -> bool:
        return bool(self._pk)

    @property
    def can_decrypt(self) -> bool:
        return bool(self._sk)

    @property
    def can_exchange_key(self) -> bool:
        return bool(self._sk and self._uid)

    def generate_pk(self, sk: bytes) -> bytes:
        """Generate public key from secret key."""

        return point_to_bytes(self._core.generate_pk(bytes_to_int(sk)), self._pc_mode)

    def generate_keypair(self) -> Tuple[bytes, bytes]:
        """Generate key pair.

        Returns:
            bytes: Secret key
            bytes: Public key.
        """

        sk, pk = self._core.generate_keypair()
        return int_to_bytes(sk), point_to_bytes(pk, self._pc_mode)

    def verify_pk(self, pk: bytes) -> bool:
        """Verify if a public key is valid.

        Args:
            pk (bytes): public key.

        Returns:
            bool: Whether valid.
        """

        return self._core.verify_pk(bytes_to_point(pk))

    def sign(self, message: bytes) -> Tuple[bytes, bytes]:
        """Generate signature on message.

        Returns:
            bytes: r
            bytes: s
        """

        if not self.can_sign:
            raise errors.RequireArgumentError("sign", "sk", "ID")

        r, s = self._core.sign(message, self._sk, self._uid, self._pk)
        return int_to_bytes(r), int_to_bytes(s)

    def verify(self, message: bytes, r: bytes, s: bytes) -> bool:
        """Verify a message and it's signature."""

        if not self.can_verify:
            raise errors.RequireArgumentError("verify", "pk", "ID")

        return self._core.verify(message, bytes_to_int(r), bytes_to_int(s), self._uid, self._pk)

    def encrypt(self, plain: bytes) -> bytes:
        """Encrypt.

        Args:
            plain (bytes): plain text to be encrypted.
        """

        if not self.can_encrypt:
            raise errors.RequireArgumentError("encrypt", "pk")

        C1, C2, C3 = self._core.encrypt(plain, self._pk)

        cipher = bytearray()
        cipher.extend(point_to_bytes(C1, self._pc_mode))
        cipher.extend(C3)
        cipher.extend(C2)

        return bytes(cipher)

    def decrypt(self, cipher: bytes) -> bytes:
        """Decrypt.

        Raises:
            ValueError: Invalid PC byte.
        """

        if not self.can_decrypt:
            raise errors.RequireArgumentError("decrypt", "sk")

        length = self._core.ecdlp.fp.e_length
        mode = cipher[0]
        if mode == 0x04 or mode == 0x06 or mode == 0x07:
            C1 = cipher[:1 + length * 2]
            c1_length = 1 + length * 2
        elif mode == 0x02 or mode == 0x03:
            C1 = cipher[:1 + length]
            c1_length = 1 + length
        else:
            raise errors.InvalidPCError(mode)

        hash_length = self._core._hash_cls.hash_length()
        C3 = cipher[c1_length:c1_length + hash_length]
        C2 = cipher[c1_length + hash_length:]

        return self._core.decrypt(bytes_to_point(C1), C2, C3, self._sk)

    def begin_key_exchange(self) -> Tuple[bytes, int]:
        """Begin key exchange.

        Returns:
            bytes: Random point, will be sent to another user.
            int: t, will be used in next step.
        """

        if not self.can_exchange_key:
            raise errors.RequireArgumentError("key exchange", "sk", "ID")

        R, t = self._core.begin_key_exchange(self._sk)
        return point_to_bytes(R, self._pc_mode), t

    def end_key_exchange(self, klen: int, t: int, R: bytes, uid: bytes, pk: bytes, mode: KEYXCHG_MODE) -> bytes:
        """End key exchange and get the secret key bytes.

        Args:
            klen (int): Length of secret key in bytes to generate.
            t (int): t value of self.
            R (bytes): Random point from another user.
            uid (bytes): ID from another user.
            pk (bytes): Public key from another user.
            mode (KEYXCHG_MODE): Key exchange mode, initiator or responder.

        Returns:
            bytes: Secret key of klen bytes.
        """

        R = bytes_to_point(R)
        pk = bytes_to_point(pk)
        S = self._core.get_secret_point(t, R, pk)

        if mode is KEYXCHG_MODE.INITIATOR:
            return self._core.generate_skey(klen, S, self._uid, self._pk, uid, pk)
        elif mode is KEYXCHG_MODE.RESPONDER:
            return self._core.generate_skey(klen, S, uid, pk, self._uid, self._pk)
        else:
            raise TypeError(f"Invalid key exchange mode: {mode}")



"""Prime field operations."""

from typing import Tuple, Union

from . import errors

__all__ = [
    "Fp2Ele",
    "Fp4Ele",
    "Fp12Ele",
    "FpExEle",
    "PrimeFieldBase",
    "PrimeField",
    "PrimeField2",
    "PrimeField4",
    "PrimeField12",
]

# towering method: 1-2-4-12
# [11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0]
# [[[11, 5], [8, 2]], [[10, 4], [7, 1]], [[9, 3], [6, 0]]]

Fp2Ele = Tuple[int, int]
Fp4Ele = Tuple[Fp2Ele, Fp2Ele]
Fp12Ele = Tuple[Fp4Ele, Fp4Ele, Fp4Ele]
FpExEle = Union[int, Fp2Ele, Fp4Ele, Fp12Ele]


class PrimeFieldBase:
    """Base class of Fp operations."""

    @classmethod
    def zero(cls) -> FpExEle:
        """Get Zero."""
        raise NotImplementedError

    @classmethod
    def one(cls) -> FpExEle:
        """Get One."""
        raise NotImplementedError

    @classmethod
    def iszero(cls, x: FpExEle) -> bool:
        raise NotImplementedError

    @classmethod
    def isone(cls, x: FpExEle) -> bool:
        raise NotImplementedError

    @classmethod
    def extend(cls, x: FpExEle) -> FpExEle:
        """Extend element."""
        raise NotImplementedError

    def __init__(self, p: int) -> None:
        raise NotImplementedError

    def isoppo(self, x: FpExEle, y: FpExEle) -> bool:
        """Whether is opposite."""
        raise NotImplementedError

    def neg(self, x: FpExEle) -> FpExEle:
        raise NotImplementedError

    def sadd(self, n: int, x: FpExEle) -> FpExEle:
        """Scalar add."""
        raise NotImplementedError

    def smul(self, k: int, x: FpExEle) -> FpExEle:
        """Scalar mul."""
        raise NotImplementedError

    def pmul(self, x: FpExEle, y: FpExEle) -> FpExEle:
        """Multiply by position."""
        raise NotImplementedError

    def add(self, x: FpExEle, y: FpExEle) -> FpExEle:
        raise NotImplementedError

    def sub(self, x: FpExEle, y: FpExEle) -> FpExEle:
        raise NotImplementedError

    def mul(self, x: FpExEle, y: FpExEle) -> FpExEle:
        raise NotImplementedError

    def inv(self, x: FpExEle) -> FpExEle:
        raise NotImplementedError

    def pow(self, x: FpExEle, e: int) -> FpExEle:
        raise NotImplementedError

    def sqrt(self, x: FpExEle) -> Union[FpExEle, None]:
        raise NotImplementedError

    def frob(self, x: FpExEle) -> FpExEle:
        """Frobenius."""
        raise NotImplementedError

    def etob(self, e: FpExEle) -> bytes:
        """Convert domain element to bytes."""
        raise NotImplementedError

    def btoe(self, b: bytes) -> FpExEle:
        """Convert bytes to domain element."""
        raise NotImplementedError


class PrimeField(PrimeFieldBase):
    """Fp operations."""

    _ZERO = 0
    _ONE = 1

    @classmethod
    def zero(cls) -> int:
        return cls._ZERO

    @classmethod
    def one(cls) -> int:
        return cls._ONE

    @classmethod
    def iszero(cls, x: int) -> bool:
        return x == cls._ZERO

    @classmethod
    def isone(cls, x: int) -> bool:
        return x == cls._ONE

    @classmethod
    def extend(cls, x: int) -> int:
        return x

    def __init__(self, p: int) -> None:
        self.p = p
        self.p_bitlength = self.p.bit_length()
        self.p_length = (self.p_bitlength + 7) >> 3
        self.e_length = self.p_length

        self._u, self._r = divmod(self.p, 8)

        if self._r == 1:
            self.sqrt = self._sqrt_8u1
        elif self._r == 3:
            self._u = self._u * 2
            self.sqrt = self._sqrt_4u3
        elif self._r == 5:
            self.sqrt = self._sqrt_8u5
        elif self._r == 7:
            self._u = self._u * 2 + 1
            self.sqrt = self._sqrt_4u3
        else:
            raise errors.InvalidArgumentError(f"0x{p:x} is not a prime number.")

    def isoppo(self, x: int, y: int) -> bool:
        return x == 0 and y == 0 or x + y == self.p

    def neg(self, x: int) -> int:
        return self.p - x if x > 0 else x

    def sadd(self, n: int, x: int) -> int:
        return (n + x) % self.p

    def smul(self, k: int, x: int) -> int:
        return (k * x) % self.p

    def pmul(self, x: int, y: int) -> int:
        return (x * y) % self.p

    def add(self, x: int, y: int) -> int:
        return (x + y) % self.p

    def sub(self, x: int, y: int) -> int:
        return (x - y) % self.p

    def mul(self, x: int, y: int) -> int:
        return (x * y) % self.p

    def inv(self, x: int):
        """Modular inverse."""

        r1 = self.p
        r2 = x
        t1 = 0
        t2 = 1
        while r2 > 0:
            q, r = divmod(r1, r2)
            r1 = r2
            r2 = r
            t = t1 - q * t2
            t1 = t2
            t2 = t
        return t1 % self.p

    def pow(self, x: int, e: int) -> int:
        return pow(x, e, self.p)

    def _lucas(self, X: int, Y: int, k: int) -> Tuple[int, int]:
        """Lucas sequence, k begin at 0.

        Uk = X * Uk-1 - Y * Uk-2
        Vk = X * Vk-1 - Y * Vk-2

        Returns:
            (int, int): The k-th lucas value pair.
        """

        p = self.p

        delta = (X * X - 4 * Y) % p
        inv2 = self.inv(2)

        U, V = 0, 2
        for i in f"{k:b}":
            U, V = (U * V) % p, ((V * V + delta * U * U) * inv2) % p
            if i == "1":
                U, V = ((X * U + V) * inv2) % p, ((X * V + delta * U) * inv2) % p
        return U, V

    def _sqrt_4u3(self, x: int) -> Union[int, None]:
        """sqrt_8u3 and sqrt_8u7"""
        p = self.p
        u = self._u

        y = pow(x, u + 1, p)
        if (y * y) % p == x:
            return y
        return None

    def _sqrt_8u5(self, x: int) -> Union[int, None]:
        p = self.p
        u = self._u

        z = pow(x, 2 * u + 1, p)
        if z == 1:
            return pow(x, u + 1, p)
        if z == p - 1:
            return (2 * x * pow(4 * x, u, p)) % p
        return None

    def _sqrt_8u1(self, x: int) -> Union[int, None]:
        p = self.p
        p_1 = p - 1
        _4u1 = 4 * self._u + 1
        inv2 = self.inv(2)

        Y = x
        for X in range(1, p):
            U, V = self._lucas(X, Y, _4u1)

            if (V * V - 4 * Y) % p == 0:
                return (V * inv2) % p

            if U != 1 or U != p_1:
                return None

        return None

    def sqrt(self, x: int) -> Union[int, None]:
        """Square root."""
        raise NotImplementedError

    def frob(self, x: int) -> int:
        return x

    def etob(self, e: int) -> bytes:
        return e.to_bytes(self.e_length, "big")

    def btoe(self, b: bytes) -> int:
        return int.from_bytes(b, "big")


class PrimeField2(PrimeFieldBase):
    """Fp2 operations."""

    _ALPHA: int = -2

    _ZERO = (PrimeField.zero(), PrimeField.zero())
    _ONE = (PrimeField.zero(), PrimeField.one())

    @classmethod
    def extend(cls, x: Union[int, Fp2Ele]) -> Fp2Ele:
        if isinstance(x, int):
            return (PrimeField.zero(), x)
        return x

    @classmethod
    def zero(cls) -> Fp2Ele:
        return cls._ZERO

    @classmethod
    def one(cls) -> Fp2Ele:
        return cls._ONE

    @classmethod
    def iszero(cls, X: Fp2Ele) -> bool:
        return X == cls._ZERO

    @classmethod
    def isone(cls, X: Fp2Ele) -> bool:
        return X == cls._ONE

    def __init__(self, p: int) -> None:
        self.fp = PrimeField(p)
        self.e_length = self.fp.e_length * 2

        self._f1 = self.pow((self.fp.one(), self.fp.zero()), p)

    def isoppo(self, X: Fp2Ele, Y: Fp2Ele) -> bool:
        return all(self.fp.isoppo(i1, i2) for i1, i2 in zip(X, Y))

    def neg(self, X: Fp2Ele) -> Fp2Ele:
        return tuple(self.fp.neg(i) for i in X)

    def sadd(self, n: int, x: Fp2Ele) -> Fp2Ele:
        x = list(x)
        x[-1] = self.fp.sadd(n, x[-1])
        return tuple(x)

    def smul(self, k: int, x: Fp2Ele) -> Fp2Ele:
        return tuple(self.fp.smul(k, i) for i in x)

    def pmul(self, X: Fp2Ele, Y: Fp2Ele) -> Fp2Ele:
        return tuple(self.fp.pmul(i1, i2) for i1, i2 in zip(X, Y))

    def add(self, X: Fp2Ele, Y: Fp2Ele) -> Fp2Ele:
        return tuple(self.fp.add(i1, i2) for i1, i2 in zip(X, Y))

    def sub(self, X: Fp2Ele, Y: Fp2Ele) -> Fp2Ele:
        return tuple(self.fp.sub(i1, i2) for i1, i2 in zip(X, Y))

    def mul(self, X: Fp2Ele, Y: Fp2Ele) -> Fp2Ele:
        a = self.fp.add
        s = self.fp.sub
        m = self.fp.mul

        X1, X0 = X
        Y1, Y0 = Y
        U = self._ALPHA

        X1mY1 = m(X1, Y1)
        X0mY0 = m(X0, Y0)

        X1aX0_m_Y1aY0 = m(a(X1, X0), a(Y1, Y0))
        Z1 = s(X1aX0_m_Y1aY0, a(X1mY1, X0mY0))
        Z0 = a(m(U, X1mY1), X0mY0)

        return Z1, Z0

    def inv(self, X: Fp2Ele) -> Fp2Ele:
        n = self.fp.neg
        s = self.fp.sub
        m = self.fp.mul

        X1, X0 = X
        U = self._ALPHA

        UmX1mX1_s_X0mX0 = s(m(U, m(X1, X1)), m(X0, X0))
        invdet = self.fp.inv(UmX1mX1_s_X0mX0)

        Y1 = m(X1, invdet)
        Y0 = m(n(X0), invdet)

        return Y1, Y0

    def conj(self, X: Fp2Ele) -> Fp2Ele:
        x1, x0 = X
        return (self.fp.neg(x1), x0)

    def pow(self, X: Fp2Ele, e: int) -> Fp2Ele:
        Y = X
        for i in f"{e:b}"[1:]:
            Y = self.mul(Y, Y)
            if i == "1":
                Y = self.mul(Y, X)
        return Y

    def sqrt(self, X: Fp2Ele) -> Union[Fp2Ele, None]:
        n = self.fp.neg
        a = self.fp.add
        s = self.fp.sub
        m = self.fp.mul

        X1, X0 = X
        U = s(m(X0, X0), m(n(self._ALPHA), m(X1, X1)))

        w1 = self.fp.sqrt(U)
        if w1 is None:
            return None

        w2 = n(w1)
        i2 = self.fp.inv(2)

        for w in (w1, w2):
            V = m(a(X0, w), i2)
            y = self.fp.sqrt(V)
            if y is None:
                continue

            Y1 = m(X1, self.fp.inv(m(2, y)))
            Y0 = y

            return Y1, Y0

        return None

    def frob(self, X: Fp2Ele) -> Fp2Ele:
        f = self.fp.frob
        X1, X0 = X
        X1 = self.mul(self.extend(f(X1)), self._f1)
        X0 = self.extend(f(X0))
        return self.add(X1, X0)

    def etob(self, e: Fp2Ele) -> bytes:
        b = bytearray()
        for i in e:
            b.extend(self.fp.etob(i))
        return bytes(b)

    def btoe(self, b: bytes) -> Fp2Ele:
        len_ = self.fp.e_length
        return tuple(self.fp.btoe(b[i:i+len_]) for i in range(0, len(b), len_))


class PrimeField4(PrimeFieldBase):
    """Fp4 operations."""

    _ALPHA: Fp2Ele = (1, 0)
    _ZERO = (PrimeField2.zero(), PrimeField2.zero())
    _ONE = (PrimeField2.zero(), PrimeField2.one())

    @classmethod
    def extend(cls, x: Union[int, Fp2Ele, Fp4Ele]) -> Fp4Ele:
        if isinstance(x, int):
            return (PrimeField2.zero(), (PrimeField.zero(), x))
        elif isinstance(x[0], int):
            return (PrimeField2.zero(), x)
        return x

    @classmethod
    def zero(cls) -> Fp4Ele:
        return cls._ZERO

    @classmethod
    def one(cls) -> Fp4Ele:
        return cls._ONE

    @classmethod
    def iszero(cls, X: Fp4Ele) -> bool:
        return X == cls._ZERO

    @classmethod
    def isone(cls, X: Fp4Ele) -> bool:
        return X == cls._ONE

    def __init__(self, p: int) -> None:
        self.fp2 = PrimeField2(p)
        self.e_length = self.fp2.e_length * 2

        self._f1 = self.pow((self.fp2.one(), self.fp2.zero()), p)

    def isoppo(self, X: Fp4Ele, Y: Fp4Ele) -> bool:
        return all(self.fp2.isoppo(i1, i2) for i1, i2 in zip(X, Y))

    def neg(self, X: Fp4Ele) -> Fp4Ele:
        return tuple(self.fp2.neg(i) for i in X)

    def sadd(self, n: int, x: Fp4Ele) -> Fp4Ele:
        x = list(x)
        x[-1] = self.fp2.sadd(n, x[-1])
        return tuple(x)

    def smul(self, k: int, x: Fp4Ele) -> Fp4Ele:
        return tuple(self.fp2.smul(k, i) for i in x)

    def pmul(self, X: Fp4Ele, Y: Fp4Ele) -> Fp4Ele:
        return tuple(self.fp2.pmul(i1, i2) for i1, i2 in zip(X, Y))

    def add(self, X: Fp4Ele, Y: Fp4Ele) -> Fp4Ele:
        return tuple(self.fp2.add(i1, i2) for i1, i2 in zip(X, Y))

    def sub(self, X: Fp4Ele, Y: Fp4Ele) -> Fp4Ele:
        return tuple(self.fp2.sub(i1, i2) for i1, i2 in zip(X, Y))

    def mul(self, X: Fp4Ele, Y: Fp4Ele) -> Fp4Ele:
        a = self.fp2.add
        s = self.fp2.sub
        m = self.fp2.mul

        X1, X0 = X
        Y1, Y0 = Y
        U = self._ALPHA

        X1mY1 = m(X1, Y1)
        X0mY0 = m(X0, Y0)

        X1aX0_m_Y1aY0 = m(a(X1, X0), a(Y1, Y0))
        Z1 = s(X1aX0_m_Y1aY0, a(X1mY1, X0mY0))
        Z0 = a(m(U, X1mY1), X0mY0)

        return Z1, Z0

    def inv(self, X: Fp4Ele) -> Fp4Ele:
        n = self.fp2.neg
        s = self.fp2.sub
        m = self.fp2.mul

        X1, X0 = X
        U = self._ALPHA

        UmX1mX1_s_X0mX0 = s(m(U, m(X1, X1)), m(X0, X0))
        invdet = self.fp2.inv(UmX1mX1_s_X0mX0)

        Y1 = m(X1, invdet)
        Y0 = m(n(X0), invdet)

        return Y1, Y0

    def conj(self, X: Fp4Ele) -> Fp4Ele:
        X1, X0 = X
        return (self.fp2.neg(X1), X0)

    def pow(self, X: Fp4Ele, e: int) -> Fp4Ele:
        Y = X
        for i in f"{e:b}"[1:]:
            Y = self.mul(Y, Y)
            if i == "1":
                Y = self.mul(Y, X)
        return Y

    def sqrt(self, X: Fp4Ele) -> Union[Fp4Ele, None]:
        raise NotImplementedError

    def frob(self, X: Fp4Ele) -> Fp4Ele:
        f = self.fp2.frob
        X1, X0 = X
        X1 = self.mul(self.extend(f(X1)), self._f1)
        X0 = self.extend(f(X0))
        return self.add(X1, X0)

    def etob(self, e: Fp4Ele) -> bytes:
        b = bytearray()
        for i in e:
            b.extend(self.fp2.etob(i))
        return bytes(b)

    def btoe(self, b: bytes) -> Fp4Ele:
        len_ = self.fp2.e_length
        return tuple(self.fp2.btoe(b[i:i+len_]) for i in range(0, len(b), len_))


class PrimeField12(PrimeFieldBase):
    """Fp12 operations."""

    _ALPHA: Fp4Ele = ((0, 1), (0, 0))
    _ZERO = (PrimeField4.zero(), PrimeField4.zero(), PrimeField4.zero())
    _ONE = (PrimeField4.zero(), PrimeField4.zero(), PrimeField4.one())

    @classmethod
    def extend(cls, x: Union[int, Fp2Ele, Fp4Ele, Fp12Ele]) -> Fp12Ele:
        if isinstance(x, int):
            return (PrimeField4.zero(), PrimeField4.zero(), (PrimeField2.zero(), (PrimeField.zero(), x)))
        elif isinstance(x[0], int):
            return (PrimeField4.zero(), PrimeField4.zero(), (PrimeField2.zero(), x))
        elif isinstance(x[0][0], int):
            return (PrimeField4.zero(), PrimeField4.zero(), x)
        return x

    @classmethod
    def zero(cls) -> Fp12Ele:
        return cls._ZERO

    @classmethod
    def one(cls) -> Fp12Ele:
        return cls._ONE

    @classmethod
    def iszero(cls, X: Fp12Ele) -> bool:
        return X == cls._ZERO

    @classmethod
    def isone(cls, X: Fp12Ele) -> bool:
        return X == cls._ONE

    def __init__(self, p: int) -> None:
        self.fp4 = PrimeField4(p)
        self.e_length = self.fp4.e_length * 3

        self._f1 = self.pow((self.fp4.zero(), self.fp4.one(), self.fp4.zero()), p)
        self._f2 = self.mul(self._f1, self._f1)

        self._frob1_factor = self.frob((((1, 1), (1, 1)), ((1, 1), (1, 1)), ((1, 1), (1, 1))))
        self._frob2_factor = self.frob1(self._frob1_factor)
        self._frob3_factor = self.frob1(self._frob2_factor)
        self._frob6_factor = self.frob3(self._frob3_factor)

    def isoppo(self, X: Fp12Ele, Y: Fp12Ele) -> bool:
        return all(self.fp4.isoppo(i1, i2) for i1, i2 in zip(X, Y))

    def neg(self, X: Fp12Ele) -> Fp12Ele:
        return tuple(self.fp4.neg(i) for i in X)

    def sadd(self, n: int, x: Fp12Ele) -> Fp12Ele:
        x = list(x)
        x[-1] = self.fp4.sadd(n, x[-1])
        return tuple(x)

    def smul(self, k: int, x: Fp12Ele) -> Fp12Ele:
        return tuple(self.fp4.smul(k, i) for i in x)

    def pmul(self, X: Fp12Ele, Y: Fp12Ele) -> Fp12Ele:
        return tuple(self.fp4.pmul(i1, i2) for i1, i2 in zip(X, Y))

    def add(self, X: Fp12Ele, Y: Fp12Ele) -> Fp12Ele:
        return tuple(self.fp4.add(i1, i2) for i1, i2 in zip(X, Y))

    def sub(self, X: Fp12Ele, Y: Fp12Ele) -> Fp12Ele:
        return tuple(self.fp4.sub(i1, i2) for i1, i2 in zip(X, Y))

    def mul(self, X: Fp12Ele, Y: Fp12Ele) -> Fp12Ele:
        a = self.fp4.add
        s = self.fp4.sub
        m = self.fp4.mul

        X2, X1, X0 = X
        Y2, Y1, Y0 = Y
        U = self._ALPHA

        X2mY2, X1mY1, X0mY0 = m(X2, Y2), m(X1, Y1), m(X0, Y0)
        X2aX1, X2aX0, X1aX0 = a(X2, X1), a(X2, X0), a(X1, X0)
        Y2aY1, Y2aY0, Y1aY0 = a(Y2, Y1), a(Y2, Y0), a(Y1, Y0)

        X2aX1_m_Y2aY1 = m(X2aX1, Y2aY1)
        X2aX0_m_Y2aY0 = m(X2aX0, Y2aY0)
        X1aX0_m_Y1aY0 = m(X1aX0, Y1aY0)

        UmX2mY2 = m(U, X2mY2)
        X2mY1_a_X1Y2 = s(X2aX1_m_Y2aY1, a(X2mY2, X1mY1))

        Z2 = s(a(X2aX0_m_Y2aY0, X1mY1), a(X2mY2, X0mY0))
        Z1 = s(a(UmX2mY2, X1aX0_m_Y1aY0), a(X1mY1, X0mY0))
        Z0 = a(m(U, X2mY1_a_X1Y2), X0mY0)

        return Z2, Z1, Z0

    def inv(self, X: Fp12Ele) -> Fp12Ele:
        a = self.fp4.add
        s = self.fp4.sub
        m = self.fp4.mul

        X2, X1, X0 = X
        U = self._ALPHA

        UmX2 = m(U, X2)
        UmX1 = m(U, X1)

        X1mX1_s_X2mX0 = s(m(X1, X1), m(X2, X0))
        UmX2mX2_s_X1X0 = s(m(UmX2, X2), m(X1, X0))
        X0mX0_s_UmX2mX1 = s(m(X0, X0), m(UmX2, X1))

        det = a(m(UmX2, UmX2mX2_s_X1X0), a(m(UmX1, X1mX1_s_X2mX0), m(X0, X0mX0_s_UmX2mX1)))
        invdet = self.fp4.inv(det)

        Y2 = m(X1mX1_s_X2mX0, invdet)
        Y1 = m(UmX2mX2_s_X1X0, invdet)
        Y0 = m(X0mX0_s_UmX2mX1, invdet)

        return Y2, Y1, Y0

    def pow(self, X: Fp12Ele, e: int) -> Fp12Ele:
        Y = X
        for i in f"{e:b}"[1:]:
            Y = self.mul(Y, Y)
            if i == "1":
                Y = self.mul(Y, X)
        return Y

    def sqrt(self, X: Fp12Ele) -> Union[Fp12Ele, None]:
        raise NotImplementedError

    def frob(self, X: Fp12Ele) -> Fp12Ele:
        f = self.fp4.frob
        X2, X1, X0 = X
        X2 = self.mul(self.extend(f(X2)), self._f2)
        X1 = self.mul(self.extend(f(X1)), self._f1)
        X0 = self.extend(f(X0))
        return self.add(self.add(X2, X1), X0)

    def frob1(self, X: Fp12Ele) -> Fp12Ele:
        return self.pmul(X, self._frob1_factor)

    def frob2(self, X: Fp12Ele) -> Fp12Ele:
        return self.pmul(X, self._frob2_factor)

    def frob3(self, X: Fp12Ele) -> Fp12Ele:
        return self.pmul(X, self._frob3_factor)

    def frob6(self, X: Fp12Ele) -> Fp12Ele:
        return self.pmul(X, self._frob6_factor)

    def etob(self, e: Fp12Ele) -> bytes:
        b = bytearray()
        for i in e:
            b.extend(self.fp4.etob(i))
        return bytes(b)

    def btoe(self, b: bytes) -> Fp12Ele:
        len_ = self.fp4.e_length
        return tuple(self.fp4.btoe(b[i:i+len_]) for i in range(0, len(b), len_))
